import pytest

from panel.template.base import BasicTemplate

from lumen.dashboard import Config


class CustomTemplate(BasicTemplate):
    pass

def no_arg_fn():
    return

def one_arg_fn(arg):
    return

def test_config_template():
    config = Config.from_spec({'template': 'lumen.tests.test_config.CustomTemplate'})
    assert config.template is CustomTemplate

def test_config_callbacks():
    config = Config.from_spec({
        'on_session_created': 'lumen.tests.test_config.no_arg_fn',
        'on_session_destroyed': 'lumen.tests.test_config.one_arg_fn',
        'on_loaded': 'lumen.tests.test_config.no_arg_fn',
        'on_error': 'lumen.tests.test_config.one_arg_fn',
        'on_update': 'lumen.tests.test_config.one_arg_fn',
    })
    assert config.on_session_created is no_arg_fn
    assert config.on_session_destroyed is one_arg_fn
    assert config.on_loaded is no_arg_fn
    assert config.on_error is one_arg_fn
    assert config.on_update is one_arg_fn

def test_config_callback_validation():
    with pytest.raises(ValueError, match=r'on_session_created callback must have signature func\(\), got func\(arg\)') as e:
        Config.from_spec({
            'on_session_created': 'lumen.tests.test_config.one_arg_fn',
        })
        assert e is None



from __future__ import annotations

import os
import pathlib

from typing import ClassVar, Type

import param  # type: ignore
import yaml

from panel.reactive import ReactiveHTML

from ..util import expand_spec
from .state import state


class Editor(ReactiveHTML):

    __abstract = True


class GalleryItem(ReactiveHTML):

    description = param.String(default="", doc="""
        A description for the gallery component.""")

    spec = param.Dict(default={}, precedence=-1, doc="""
        The dictionary specification of the gallery item.""")

    selected = param.Boolean(default=False, doc="""
        Whether the item has been selected.""")

    view = param.Parameter(doc="""
        A Panel view of the contents of the gallery item.""")

    margin = param.Integer(default=0)

    thumbnail = param.Filename(precedence=-1)

    sizing_mode= param.String(default='stretch_both')

    __abstract = True

    _template = """
    <span style="font-size: 1.2em; font-weight: bold;">{{ name }}</p>
    <fast-switch id="selected" checked=${selected} style="float: right"></fast-switch>
    <div id="details" style="margin: 1em 0;">
      ${view}
    </div>
    <p style="height: 4em; max-width: 320px;">{{ description }}</p>
    <fast-button id="edit-button" style="width: 320px;" onclick="${_open_modal}">Edit</fast-button>
    """

    def _open_modal(self, event):
        if state.modal.objects == self._modal_content:
            state.template.open_modal()
            return
        state.modal.loading = True
        state.modal[:] = self._modal_content
        state.modal.loading = False
        state.template.open_modal()


class Gallery(ReactiveHTML):

    path = param.Foldername()

    items = param.Dict(default={})

    sizing_mode = param.String(default='stretch_width', readonly=True)

    hidden = param.Boolean(default=True)

    _editor_type: ClassVar[Type[Editor] | None] = None

    _gallery_item = GalleryItem

    _glob_pattern = '*.y*ml'

    __abstract = True

    def __init__(self, **params):
        path = params.get('path', self.path)
        components = pathlib.Path(path).glob(self._glob_pattern)
        params['items'] = items = {}
        for component in components:
            with open(component, encoding='utf-8') as f:
                yaml_spec = f.read()
                spec = yaml.safe_load(expand_spec(yaml_spec))
            if not spec:
                continue
            metadata = spec.pop('metadata', {})
            if 'name' in metadata:
                name = metadata['name']
            else:
                name = '.'.join(component.name.split('.')[:-1])
            if 'thumbnail' in metadata:
                thumbnail = metadata['thumbnail']
                if not pathlib.Path(thumbnail).is_absolute():
                    thumbnail = path / pathlib.Path(thumbnail)
            else:
                thumbnail = '.'.join(str(component).split('.')[:-1]) + '.png'
            kwargs = {'name': name, 'spec': spec}
            if os.path.isfile(thumbnail):
                kwargs['thumbnail'] = thumbnail
            if self._editor_type:
                kwargs['editor'] = self._editor_type(**kwargs)
            if 'description' in metadata:
                kwargs['description'] = metadata['description']
            kwargs = self._preprocess_kwargs(kwargs)
            items[name] = item = self._gallery_item(**kwargs)
            item.param.watch(self._selected, ['selected'])
        super().__init__(**params)

    def _preprocess_kwargs(self, kwargs):
        return kwargs

    def _selected(self, event):
        """
        Called when a GalleryItem is selected
        """

    def _open_modal(self, event):
        if state.modal.objects == self._modal_content:
            state.template.open_modal()
            return
        state.modal.loading = True
        state.template.open_modal()
        state.modal[:] = self._modal_content
        state.modal.loading = False



"""
The Transform components allow transforming tables in arbitrary ways.
"""
from __future__ import annotations

import datetime as dt
import hashlib

from typing import (
    TYPE_CHECKING, Any, Callable, ClassVar, Dict, List, Literal, Tuple, Union,
)

import numpy as np
import pandas as pd
import panel as pn
import param  # type: ignore

from packaging.version import Version
from panel.io.cache import _generate_hash

from ..base import MultiTypeComponent
from ..state import state
from ..util import is_ref

pd_version = Version(pd.__version__)

if TYPE_CHECKING:
    from dask.dataframe import DataFrame as dDataFrame, Series as dSeries
    from panel.viewable import Viewable
    DataFrame = Union[pd.DataFrame, dDataFrame]
    Series = Union[pd.Series, dSeries]


class Transform(MultiTypeComponent):
    """
    `Transform` components implement transforms of `DataFrame` objects.
    """

    controls = param.List(default=[], doc="""
        Parameters that should be exposed as widgets in the UI.""")

    transform_type: ClassVar[str | None] = None

    _field_params: ClassVar[List[str]] = []

    _valid_keys: ClassVar[List[str] | Literal['params'] | None] = 'params'

    __abstract = True

    @classmethod
    def from_spec(cls, spec: Dict[str, Any] | str) -> 'Transform':
        """
        Resolves a Transform specification.

        Parameters
        ----------
        spec: dict
            Specification declared as a dictionary of parameter values.

        Returns
        -------
        The resolved Transform object.
        """
        if isinstance(spec, str):
            raise ValueError(
                "Transform cannot be materialized by reference. Please pass "
                "full specification for the transform."
            )
        spec = dict(spec)
        transform_type = Transform._get_type(spec.pop('type', None))
        new_spec, refs = {}, {}
        for k, v in spec.items():
            if is_ref(v):
                refs[k] = v
                v = state.resolve_reference(v)
            elif isinstance(v, dict):
                resolved = {}
                for sk, sv in v.items():
                    if is_ref(sv):
                        refs[f'{k}.{sk}'] = sv
                        sv = state.resolve_reference(sv)
                    resolved[sk] = sv
                v = resolved
            if (k in transform_type.param and
                isinstance(transform_type.param[k], param.ListSelector) and
                not isinstance(v, list)):
                v = [v]
            new_spec[k] = v

        # Resolve any specs for the controls
        controls, control_kwargs = [], {}
        for control in new_spec.get('controls', []):
            if isinstance(control, dict):
                ckws = {}
                if 'options' in control:
                    options = control['options']
                    if isinstance(options, str):
                        options = state.resolve_reference(options)
                    ckws['objects'] = options
                if 'start' in control or 'end' in control:
                    ckws['bounds'] = (control.get('start'), control.get('end'))
                control = control['name']
                control_kwargs[control] = ckws
            controls.append(control)
        new_spec['controls'] = controls

        # Instantiate the transform
        transform = transform_type(refs=refs, **new_spec)

        # Modify the parameters for the controls
        for p, vs in control_kwargs.items():
            p = transform.param[p]
            for attr, val in vs.items():
                if hasattr(p, attr):
                    setattr(p, attr, val)
                else:
                    attr = 'options' if attr == 'objects' else attr
                    cls.param.warning(
                        f"{transform_type.__name__} is of type {type(p).__name__} "
                        f"and has not attribute {attr!r}. Ensure the controls "
                        "parameter supports the provided options, e.g. if "
                        "you are declaring 'options' ensure that the parameter "
                        "is a param.Selector type."
                    )
        return transform

    @classmethod
    def apply_to(cls, table: DataFrame, **kwargs) -> DataFrame:
        """
        Calls the apply method based on keyword arguments passed to define transform.

        Parameters
        ----------
        table: `pandas.DataFrame`

        Returns
        -------
        A DataFrame with the results of the transformation.
        """
        return cls(**kwargs).apply(table)

    def __hash__(self) -> int:
        """
        Implements hashing to allow a Source to compute a hash key.
        """
        sha = hashlib.sha256()
        hash_vals: Tuple[Any, ...] = (type(self).__name__.encode('utf-8'),)
        hash_vals += tuple(sorted([
            (k, v) for k, v in self.param.values().items()
            if k not in Transform.param
        ]))
        sha.update(_generate_hash(hash_vals))
        return int(sha.hexdigest(), base=16)

    def apply(self, table: DataFrame) -> DataFrame:
        """
        Given a table transform it in some way and return it.

        Parameters
        ----------
        table : DataFrame
            The queried table as a DataFrame.

        Returns
        -------
        DataFrame
            A DataFrame containing the transformed data.
        """
        return table

    @property
    def control_panel(self) -> Viewable:
        return pn.Param(
            self.param, parameters=self.controls, sizing_mode='stretch_width',
            margin=(-10, 0, 5, 0)
        )

    def _drop_none_values(self, **kwargs: Dict[str, Any]) -> Dict[str, Any]:
        return {k: v for k, v in kwargs.items() if v is not None}


class Filter(Transform):
    """
    `Filter` transform implement the filtering behavior of `Filter` components.

    The filter `conditions` must be declared as a list of tuple containing
    the name of the column to be filtered and one of the following:

      * scalar: A scalar value will be matched using equality operators
      * tuple:  A tuple value specifies a numeric or date range.
      * list:   A list value specifies a set of categories to match against.
      * list(tuple): A list of tuples specifies a list of ranges.
    """

    conditions = param.List(doc="""
      List of filter conditions expressed as tuples of the column
      name and the filter value.""")

    @classmethod
    def _range_filter(cls, column: Series, start: Any, end: Any) -> Series | None:
        if column.dtype.kind == 'M':
            if isinstance(start, dt.date) and not isinstance(start, dt.datetime):
                start = dt.datetime(*start.timetuple()[:3], 0, 0, 0)
            if isinstance(end, dt.date) and not isinstance(end, dt.datetime):
                end = dt.datetime(*end.timetuple()[:3], 23, 59, 59)
        if start is None and end is None:
            return None
        elif start is None:
            mask = column<=end
        elif end is None:
            mask = column>=start
        else:
            mask = (column>=start) & (column<=end)
        return mask

    def apply(self, df: DataFrame) -> DataFrame:
        filters = []
        for k, val in self.conditions:
            if k not in df.columns:
                continue
            column = df[k]
            if np.isscalar(val) or isinstance(val, dt.date):
                if (column.dtype.kind == 'M' and isinstance(val, dt.date)
                    and not isinstance(val, dt.datetime)):
                    val = dt.datetime(*val.timetuple()[:3], 0, 0, 0)
                mask = column == val
            elif isinstance(val, list) and all(isinstance(v, tuple) and len(v) == 2 for v in val):
                val = [v for v in val if v is not None]
                if not val:
                    continue
                mask = self._range_filter(column, *val[0])
                for v in val[1:]:
                    mask |= self._range_filter(column, *v)
                if mask is not None:
                    filters.append(mask)
                continue
            elif isinstance(val, list):
                if not val:
                    continue
                mask = column.isin(val)
            elif isinstance(val, tuple):
                mask = self._range_filter(column, *val)
            else:
                self.param.warning(
                    'Condition {val!r} on {col!r} column not understood. '
                    'Filter query will not be applied.'
                )
                continue
            if mask is not None:
                filters.append(mask)
        if filters:
            mask = filters[0]
            for f in filters[1:]:
                mask &= f
            df = df[mask]
        return df


class HistoryTransform(Transform):
    """
    `HistoryTransform` accumulates a history of the queried data.

    The internal buffer accumulates data up to the supplied `length`
    and (optionally) adds a date_column to the data.
    """

    date_column = param.Selector(doc="""
        If defined adds a date column with the supplied name.""")

    length = param.Integer(default=10, bounds=(1, None), doc="""
        Accumulates a history of data.""")

    transform_type = 'history'

    _field_params = ['date_column']

    def __init__(self, **params):
        super().__init__(**params)
        self._buffer = []

    def apply(self, table: DataFrame) -> DataFrame:
        """
        Accumulates a history of the data in a buffer up to the
        declared `length` and optionally adds the current datetime to
        the declared `date_column`.

        Parameters
        ----------
        data : DataFrame
            The queried table as a DataFrame.

        Returns
        -------
        DataFrame
            A DataFrame containing the buffered history of the data.
        """
        if self.date_column:
            table = table.copy()
            table[self.date_column] = dt.datetime.now()
        self._buffer.append(table)
        self._buffer[:] = self._buffer[-self.length:]
        return pd.concat(self._buffer)


class Aggregate(Transform):
    """
    `Aggregate` one or more columns or indexes, see `pandas.DataFrame.groupby`.

    `by` must be provided.

    `df.groupby(<by>)[<columns>].<method>()[.reset_index()]`
    """

    by = param.ListSelector(doc="""
        Columns or indexes to group by.""")

    columns = param.ListSelector(allow_None=True, doc="""
        Columns to aggregate.""")

    with_index = param.Boolean(default=True, doc="""
        Whether to make the groupby columns indexes.""")

    method = param.String(default="mean", doc="""
        Name of the pandas aggregation method, e.g. max, min, count.""")

    kwargs = param.Dict(default={}, doc="""
        Keyword arguments to the aggregation method.""")

    transform_type: ClassVar[str] = 'aggregate'

    _field_params: ClassVar[List[str]] = ['by', 'columns']

    def apply(self, table: DataFrame) -> DataFrame:
        grouped = table.groupby(self.by)
        if self.columns:
            cols = self.columns
        else:
            cols = [
                c for c in table.select_dtypes(include='number').columns
                if c not in self.by
            ]
        grouped = grouped[cols]
        agg = getattr(grouped, self.method)(**self.kwargs)
        return agg if self.with_index else agg.reset_index()


class Sort(Transform):
    """
    `Sort` on one or more columns, see `pandas.DataFrame.sort_values`.

    `df.sort_values(<by>, ascending=<ascending>)`
    """

    by = param.ListSelector(default=[], doc="""
       Columns or indexes to sort by.""")

    ascending = param.ClassSelector(default=True, class_=(bool, list), doc="""
       Sort ascending vs. descending. Specify list for multiple sort
       orders. If this is a list of bools, must match the length of
       the by.""")

    transform_type: ClassVar[str] = 'sort'

    _field_params: ClassVar[List[str]] = ['by']

    def apply(self, table: DataFrame) -> DataFrame:
        return table.sort_values(self.by, ascending=self.ascending)


class Query(Transform):
    """
    `Query` applies the `pandas.DataFrame.query` method.

    `df.query(<query>)`
    """

    query = param.String(doc="""
        The query to apply to the table.""")

    transform_type: ClassVar[str] = 'query'

    def apply(self, table: DataFrame) -> DataFrame:
        return table.query(self.query)


class Columns(Transform):
    """
    `Columns` selects a subset of columns.

    `df[<columns>]`
    """

    columns = param.ListSelector(doc="""
        The subset of columns to select.""")

    transform_type: ClassVar[str] = 'columns'

    _field_params: ClassVar[List[str]] = ['columns']

    def apply(self, table: DataFrame) -> DataFrame:
        return table[self.columns]


class Astype(Transform):
    """
    `Astype` transforms the type of one or more columns.
    """

    dtypes = param.Dict(doc="Mapping from column name to new type.")

    transform_type: ClassVar[str] = 'as_type'

    def apply(self, table: DataFrame) -> DataFrame:
        table = table.copy()
        for col, dtype in self.dtypes.items():
            if col in table.columns:
                table[col] = table[col].astype(dtype)
        return table


class Stack(Transform):
    """
    `Stack` applies `pandas.DataFrame.stack` to the declared `level`.

    `df.stack(<level>)`
    """

    dropna = param.Boolean(default=True, doc="""
        Whether to drop rows in the resulting Frame/Series with missing values.
        Stacking a column level onto the index axis can create combinations of
        index and column values that are missing from the original
        dataframe.""")

    level = param.ClassSelector(default=-1, class_=(int, list, str), doc="""
        The indexes to stack.""")

    transform_type: ClassVar[str] = 'stack'

    def apply(self, table: DataFrame) -> DataFrame:
        return table.stack(level=self.level, dropna=self.dropna)


class Unstack(Transform):
    """
    `Unstack` applies `pandas.DataFrame.unstack` to the declared `level`.

    `df.unstack(<level>)`
    """

    fill_value = param.ClassSelector(default=None, class_=(int, str, dict), doc="""
        Replace NaN with this value if the unstack produces missing values.""")

    level = param.ClassSelector(default=-1, class_=(int, list, str), doc="""
        The indexes to unstack.""")

    transform_type: ClassVar[str] = 'unstack'

    def apply(self, table: DataFrame) -> DataFrame:
        return table.unstack(level=self.level, fill_value=self.fill_value)


class Iloc(Transform):
    """
    `Iloc` allows selecting the data with integer indexing, see `pandas.DataFrame.iloc`.

    `df.iloc[<start>:<end>]`
    """

    start = param.Integer(default=None)

    end = param.Integer(default=None)

    transform_type: ClassVar[str] = 'iloc'

    def apply(self, table: DataFrame) -> DataFrame:
        return table.iloc[self.start:self.end]


class Sample(Transform):
    """
    `Sample` returns a random sample of items.

    `df.sample(n=<n>, frac=<frac>, replace=<replace>)`
    """

    n = param.Integer(default=None, doc="""
        Number of items to return.""")

    frac = param.Number(default=None, bounds=(0, 1), doc="""
        Fraction of axis items to return.""")

    replace = param.Boolean(default=False, doc="""
        Sample with or without replacement.""")

    transform_type: ClassVar[str] = 'sample'

    def apply(self, table: DataFrame) -> DataFrame:
        return table.sample(
            **self._drop_none_values(n=self.n, frac=self.frac, replace=self.replace)
        )


class Compute(Transform):
    """
    `Compute` turns a `dask.dataframe.DataFrame` into a `pandas.DataFrame`.
    """

    transform_type: ClassVar[str] = 'compute'

    def apply(self, table: DataFrame) -> DataFrame:
        if hasattr(table, 'compute'):
            return table.compute()
        return table


class Pivot(Transform):
    """
    `Pivot` applies `pandas.DataFrame.pivot` given an index, columns, and values.
    """

    index = param.String(default=None, doc="""
        Column to use to make new frame's index.
        If None, uses existing index.""")

    columns = param.String(default=None, doc="""
        Column to use to make new frame’s columns.""")

    values = param.ListSelector(default=None, doc="""
        Column(s) to use for populating new frame’s values.
        If not specified, all remaining columns will be used
        and the result will have hierarchically indexed columns.""")

    transform_type: ClassVar[str] = 'pivot'

    def apply(self, table: DataFrame) -> DataFrame:
        pivot_table = table.pivot(
            **self._drop_none_values(
                index=self.index,
                columns=self.columns,
                values=self.values
            )
        )
        pivot_table.columns = pivot_table.columns.to_flat_index()
        return pivot_table


class Melt(Transform):
    """
    `Melt` applies the `pandas.melt` operation given the `id_vars` and `value_vars`.
    """

    id_vars = param.ListSelector(default=[], doc="""
        Column(s) to use as identifier variables.""")

    ignore_index = param.Boolean(default=True, doc="""
        If True, original index is ignored. If False, the original
        index is retained. Index labels will be repeated as
        necessary.""")

    value_vars = param.ListSelector(default=None, doc="""
        Column(s) to unpivot. If not specified, uses all columns that
        are not set as `id_vars`.""")

    var_name = param.String(default=None, doc="""
         Name to use for the 'variable' column. If None it uses
         ``frame.columns.name`` or 'variable'.""")

    value_name = param.String(default='value', doc="""
         Name to use for the 'value' column.""")

    transform_type: ClassVar[str] = 'melt'

    _field_params: ClassVar[List[str]] = ['id_vars', 'value_vars']

    def apply(self, table: DataFrame) -> DataFrame:
        melt: Callable
        if isinstance(table, pd.DataFrame):
            melt = pd.melt
        else:
            import dask.dataframe as dd
            melt = dd.melt
        return melt(
            table, id_vars=self.id_vars, value_vars=self.value_vars,
            var_name=self.var_name, value_name=self.value_name,
            ignore_index=self.ignore_index
        )


class SetIndex(Transform):
    """
    `SetIndex` promotes DataFrame columns to indexes, see `pandas.DataFrame.set_index`.

    `df.set_index(<keys>, drop=<drop>, append=<append>, verify_integrity=<verify_integrity>)`
    """

    append = param.Boolean(default=False, doc="""
        Whether to append columns to existing index.""")

    drop = param.Boolean(default=True, doc="""
        Delete columns to be used as the new index.""")

    keys = param.ClassSelector(default=None, class_=(str, list), doc="""
        This parameter can be either a single column key or a list
        containing column keys.""")

    verify_integrity = param.Boolean(default=False, doc="""
        Check the new index for duplicates. Otherwise defer the check
        until necessary. Setting to False will improve the performance
        of this method.""")

    transform_type: ClassVar[str] = 'set_index'

    _field_params: ClassVar[List[str]] = ['keys']

    def apply(self, table: DataFrame) -> DataFrame:
        return table.set_index(
            self.keys, drop=self.drop, append=self.append,
            verify_integrity=self.verify_integrity
        )


class ResetIndex(Transform):
    """
    `ResetIndex` resets DataFrame indexes to columns or drops them, see `pandas.DataFrame.reset_index`

    `df.reset_index(drop=<drop>, col_fill=<col_fill>, col_level=<col_level>, level=<level>)`
    """

    col_fill = param.String(default="", doc="""
        If the columns have multiple levels, determines how the other
        levels are named. If None then the index name is repeated.""")

    col_level = param.ClassSelector(default=0, class_=(int, str), doc="""
        If the columns have multiple levels, determines which level the
        labels are inserted into. By default it is inserted into the
        first level.""")

    drop = param.Boolean(default=False, doc="""
        Do not try to insert index into dataframe columns. This resets
        the index to the default integer index.""")

    level = param.ClassSelector(default=None, class_=(int, str, list), doc="""
        Only remove the given levels from the index. Removes all levels
        by default.""")

    transform_type: ClassVar[str] = 'reset_index'

    def apply(self, table: DataFrame) -> DataFrame:
        return table.reset_index(
            drop=self.drop, col_fill=self.col_fill, col_level=self.col_level,
            level=self.level
        )  # type: ignore


class Rename(Transform):
    """
    `Rename` renames columns or indexes, see `pandas.DataFrame.rename`.

    df.rename(mapper=<mapper>, columns=<columns>, index=<index>,
              level=<level>, axis=<axis>, copy=<copy>)
    """

    axis = param.ClassSelector(default=None, class_=(int, str), doc="""
        The axis to rename. 0 or 'index', 1 or 'columns'""")

    columns = param.Dict(default=None, doc="""
        Alternative to specifying axis (`mapper, axis=1` is equivalent to
        `columns=mapper`).""")

    copy = param.Boolean(default=False, doc="""
        Also copy underlying data.""")

    index = param.Dict(default=None, doc="""
        Alternative to specifying axis (`mapper, axis=0` is equivalent to
        `index=mapper`).""")

    mapper = param.Dict(default=None, doc="""
        Dict to apply to that axis' values. Use either `mapper` and `axis` to
        specify the axis to target with `mapper`, or `index` and `columns`.""")

    level = param.ClassSelector(default=None, class_=(int, str), doc="""
        In case of a MultiIndex, only rename labels in the specified level.""")

    transform_type: ClassVar[str] = 'rename'

    def apply(self, table: DataFrame) -> DataFrame:
        return table.rename(
            axis=self.axis, columns=self.columns, copy=self.copy,
            index=self.index, mapper=self.mapper, level=self.level,
        )  # type: ignore


class RenameAxis(Transform):
    """
    Set the name of the axis for the index or columns,
    see `pandas.DataFrame.rename_axis`.

    df.rename_axis(mapper=<mapper>, columns=<columns>, index=<index>,
                  axis=<axis>, copy=<copy>)
    """

    axis = param.ClassSelector(default=0, class_=(int, str), doc="""
        The axis to rename. 0 or 'index', 1 or 'columns'""")

    columns = param.ClassSelector(default=None, class_=(str, list, dict), doc="""
        A scalar, list-like, dict-like to apply to that axis' values.
        Note that the columns parameter is not allowed if the object
        is a Series. This parameter only apply for DataFrame type objects.
        Use either mapper and axis to specify the axis to target with
        mapper, or index and/or columns.""")

    copy = param.Boolean(default=True, doc="""
        Also copy underlying data.""")

    index = param.ClassSelector(default=None, class_=(str, list, dict), doc="""
        A scalar, list-like, dict-like to apply to that axis' values.
        Note that the columns parameter is not allowed if the object
        is a Series. This parameter only apply for DataFrame type objects.
        Use either mapper and axis to specify the axis to target with
        mapper, or index and/or columns.""")

    mapper = param.ClassSelector(default=None, class_=(str, list), doc="""
        Value to set the axis name attribute.""")

    transform_type: ClassVar[str] = 'rename_axis'

    def apply(self, table: DataFrame) -> DataFrame:
        return table.rename_axis(
            axis=self.axis, columns=self.columns, copy=self.copy,
            index=self.index, mapper=self.mapper,
        )


class Count(Transform):
    """
    Counts non-nan values in each column of the DataFrame and returns
    a new DataFrame with a single row with a count for each original
    column, see `pandas.DataFrame.count`.

    df.count(axis=<axis>, level=<level>, numeric_only=<numeric_only>).to_frame().T
    """

    axis = param.ClassSelector(default=0, class_=(int, str), doc="""
        The axis to rename. 0 or 'index', 1 or 'columns'""")

    level = param.ClassSelector(default=None, class_=(int, list, str), doc="""
        The indexes to stack.""")

    numeric_only = param.Boolean(default=False, doc="""
        Include only float, int or boolean data.""")

    transform_type: ClassVar[str] = 'count'

    def apply(self, table: DataFrame) -> DataFrame:
        kwargs = {}
        if pd_version < Version('2.0.0'):
            kwargs['level'] = self.level
        return table.count(
            axis=self.axis, numeric_only=self.numeric_only, **kwargs
        ).to_frame().T


class Sum(Transform):
    """
    Sums numeric values in each column of the DataFrame and returns a
    new DataFrame with a single row containing the sum for each
    original column, see `pandas.DataFrame.sum`.

    df.count(axis=<axis>, level=<level>).to_frame().T
    """

    axis = param.ClassSelector(default=0, class_=(int, str), doc="""
        The axis to rename. 0 or 'index', 1 or 'columns'""")

    level = param.ClassSelector(default=None, class_=(int, list, str), doc="""
        The indexes to stack.""")

    transform_type: ClassVar[str] = 'sum'

    def apply(self, table: DataFrame) -> DataFrame:
        kwargs = {}
        if pd_version < Version('2.0.0'):
            kwargs['level'] = self.level
        return table.sum(
            axis=self.axis, **kwargs
        ).to_frame().T


class Eval(Transform):
    """
    Applies an eval assignment expression to a DataFrame. The
    expression can reference columns on the original table by
    referencing `table.<column>` and must assign to a variable that
    will become a new column in the DataFrame, e.g. to divide a
    `value` column by one thousand and assign the result to a new column
    called `kilo_value` you can write an `expr` like:

        kilo_value = table.value / 1000

    See `pandas.eval` for more information.
    """

    expr = param.String(doc="""
        The expression to apply to the table.""")

    transform_type: ClassVar[str] = 'eval'

    def apply(self, table: DataFrame) -> DataFrame:
        return pd.eval(self.expr, target=table)


class DropNA(Transform):
    """
    `DropNA` drops rows with any missing values.

    `df.dropna(axis=<axis>, how=<how>, thresh=<thresh>, subset=<subset>)`
    """

    axis = param.ClassSelector(default=0, class_=(int, str), doc="""
        The axis to rename. 0 or 'index', 1 or 'columns'""")

    how = param.ObjectSelector(default='any', objects=['any', 'all'], doc="""
        Determine if row or column is removed from DataFrame, when we have
        at least one NA or all NA.""")

    thresh = param.Integer(default=None, doc="""
        Require that many non-NA values.""")

    subset = param.ListSelector(default=None, doc="""
        Labels along other axis to consider, e.g. if you are dropping rows
        these would be a list of columns to include.""")

    transform_type: ClassVar[str] = 'dropna'

    def apply(self, table: DataFrame) -> DataFrame:
        kwargs = {'axis': self.axis, 'subset': self.subset}
        if self.how:
            kwargs['how'] = self.how
        if self.thresh:
            kwargs['thresh'] = self.thresh
        return table.dropna(**kwargs)


class project_lnglat(Transform):
    """
    `project_lnglat` projects the given longitude/latitude columns to Web Mercator.

    Converts latitude and longitude values into WGS84 (Web Mercator)
    coordinates (meters East of Greenwich and meters North of the
    Equator).
    """

    longitude = param.String(default='longitude', doc="Longitude column")
    latitude = param.String(default='longitude', doc="Latitude column")

    transform_type: ClassVar[str] = 'project_lnglat'

    def apply(self, table: DataFrame) -> DataFrame:
        table = table.copy()
        longitude = table[self.longitude]
        latitude = table[self.latitude]

        origin_shift = np.pi * 6378137
        table[self.longitude] = longitude * origin_shift / 180.0
        table[self.latitude] = np.log(np.tan((90 + latitude) * np.pi / 360.0)) * origin_shift / np.pi
        return table


__all__ = [name for name, obj in locals().items() if isinstance(obj, type) and issubclass(obj, Transform)]



from __future__ import annotations

import os

from functools import partial
from typing import TYPE_CHECKING, ClassVar, List

import panel as pn
import param  # type: ignore

from panel.widgets import Widget as _PnWidget
from typing_extensions import Literal

from ..base import MultiTypeComponent
from ..state import state
from ..util import disallow_refs, is_ref, resolve_module_reference

if TYPE_CHECKING:
    from panel.viewable import Viewable

_PARAM_MAP = {
    dict : param.Dict,
    float: param.Number,
    list : param.List,
    str  : param.String
}


class Variables(param.Parameterized):
    """
    `Variables` stores a number Variable types and mirrors their
    values onto dynamically created parameters allowing other
    components to easily watch changes in a variable.
    """

    _counter: ClassVar[int] = 0

    def __init__(self, **params):
        super().__init__(**params)
        self._vars = {}
        self._watchers = {}

    def keys(self):
        return self._vars.keys()

    @classmethod
    def create_variables(cls):
        new_cls = type(f'Variables{cls._counter}', (cls,), {})()
        cls._counter += 1
        return new_cls

    @classmethod
    def from_spec(cls, spec):
        variables = cls.create_variables()
        doc = pn.state.curdoc
        if doc and doc._session_context:
            state._variables[doc] = variables
        else:
            state._variable = variables
        for name, var_spec in spec.items():
            if not isinstance(var_spec, dict):
                var_spec = {
                    'type': 'constant',
                    'default': var_spec
                }
            var = Variable.from_spec(dict(var_spec, name=name), variables)
            variables.add_variable(var)
        return variables

    def _convert_to_variable(self, var: param.Parameter | _PnWidget) -> 'Variable':
        throttled = False
        if isinstance(var, param.Parameter):
            if isinstance(var.owner, pn.widgets.Widget) and var.name in ('value', 'value_throttled'):
                throttled = var.name == 'value_throttled'
                var = var.owner
            else:
                var_type = Parameter
                var_name = var.name
                var_kwargs = {'parameter': var, 'value': getattr(var.owner, var.name)}
        else:
            var_type = None

        if isinstance(var, _PnWidget):
            var_name = var.name
            widget_type = type(var).__name__
            if not var_name:
                raise ValueError(
                    f'Cannot use {widget_type} widget without an explicit '
                    'name as a reference. Ensure any Widget passed in as '
                    'a reference has been given a name.'
                )
            var_type = Widget
            extras = {k: v for k, v in var.param.values().items() if k != 'name'}
            var_kwargs = dict(
                kind=f'{var.__module__}.{widget_type}', throttled=throttled,
                widget=var, **extras
            )

        if var_type is None:
            raise ValueError(
                f'Could not convert object of type {type(var).__name__} '
                'to a Variable. Only panel.widgets.Widget and param.Parameter '
                'types may be automatically promoted to Variables.'
            )
        var_name = var_name.replace(' ', '_')
        old_var = self._vars.get(var_name)
        if (
            isinstance(old_var, var_type) and
            ((var_type is Widget and var is old_var._widget) or
             (var_type is Parameter and var is old_var.parameter))
        ):
            return old_var
        return var_type(name=var_name, **var_kwargs)

    def add_variable(self, var: 'Variable' | _PnWidget | param.Parameter) -> 'Variable':
        """
        Adds a new variable to the Variables instance and sets up
        a parameter that can be watched.

        Arguments
        ---------
        var: Variable | panel.widgets.Widget | param.Parameter
            The Variable to add to the Variables store.
        """
        if not isinstance(var, Variable):
            var = self._convert_to_variable(var)
        old_var = self._vars.get(var.name)
        if old_var:
            if var is old_var:
                return var
            elif old_var and type(var) is type(old_var):
                vtname = type(var).__name__
                self.param.warning(
                    f'A {vtname} named {var.name!r} has already been defined and '
                    'is linked to a different source. Overriding the existing '
                    'variable and unlinking any attached components.'
                )
            else:
                self.param.warning(
                    f'A Variable named {var.name!r} of a different type has '
                    'already been defined. Overriding the existing variable '
                    'and unlinking any attached components.'
                )
            variable = self._vars.pop(var.name)
            variable.param.unwatch(self._watchers.pop(var.name))
        self._vars[var.name] = var
        self.param.add_parameter(var.name, param.Parameter(default=var.value))
        self.param.update(**{var.name: var.value})
        self._watchers[var.name] = var.param.watch(
            partial(self._update_value, var.name), 'value'
        )
        return var

    def _update_value(self, name: str, event: param.parameterized.Event):
        self.param.update({name: event.new})

    def __getitem__(self, key: str) -> 'Variable':
        if key in self.param:
            return getattr(self, key)
        else:
            raise KeyError(f'No variable named {key!r} has been defined.')

    def panel(self, variables: List[str] | None = None) -> pn.Column | None:
        if variables == []:
            return None
        column = pn.Column(name='Variables', sizing_mode='stretch_width')
        for key, var in self._vars.items():
            if variables is None or key in variables:
                var_panel = var.panel
                if var_panel is not None:
                    column.append(var_panel)
        return column


class Variable(MultiTypeComponent):
    """
    `Variable` components declare values that can be referenced from other components.

    The source of the `Variable` value can be anything from an
    environment variable to a widget or URL parameter. `Variable`
    components allow a concise way to configure other components and
    make it possible to orchestrate actions across multiple
    components.
    """

    default = param.Parameter(constant=True, doc="""
        Default value to use if no other value is defined""")

    label = param.String(default=None, doc="""
        Optional label override for variable. Used wherever the variable
        shows up in the UI.""")

    materialize = param.Boolean(default=False, constant=True, doc="""
        Whether the variable should be inlined as a constant variable
        (in the Lumen Builder).""")

    required = param.Boolean(default=False, constant=True, doc="""
        Whether the variable should be treated as required.""")

    secure = param.Boolean(default=False, constant=True, doc="""
        Whether the variable should be treated as secure.""")

    value = param.Parameter(doc="""
        The materialized value of the variable.""")

    variable_type: ClassVar[str | None] = None

    __abstract = True

    _valid_keys: ClassVar[List[str] | Literal['params'] | None] = 'params'
    _validate_params: ClassVar[bool] = True

    def __init__(self, **params):
        if 'value' not in params and 'default' in params:
            params['value'] = params['default']
        if 'label' not in params and 'name' in params:
            params['label'] = params['name']
        super().__init__(**params)
        self.param.watch(self._update_value_from_default, 'default')

    def _update_value_from_default(self, event: param.parameterized.Event):
        if event.old is self.value or self.value is self.param.value.default:
            self.value = event.new

    @classmethod
    def from_spec(cls, spec, variables=None):
        if isinstance(spec, dict):
            spec = dict(spec)
            var_type = spec.pop('type')
        else:
            var_type = 'constant'
            spec = {'default': spec}
        var_type = cls._get_type(var_type)
        resolved_spec, refs = {}, {}
        for k, val in spec.items():
            if is_ref(val):
                refs[k] = val
                val = state.resolve_reference(val, variables)
                if isinstance(val, Variable):
                    val = val.value
            resolved_spec[k] = val
        return var_type(refs=refs, **resolved_spec)

    def as_materialized(self) -> 'Constant':
        """
        If the variable is to be materialized by the builder this
        implements the conversion from a variable that references
        some external value to a materialized value.
        """
        return Constant(default=self.value)

    @property
    def panel(self) -> Viewable | None:
        """
        Optionally returns a renderable component that allows
        controlling the variable.
        """
        return None


class Constant(Variable):
    """
    `Constant` declares a constant value that can be referenced.
    """

    variable_type: ClassVar[str] = 'constant'


class EnvVariable(Variable):
    """
    `EnvVariable` fetches an environment variable that can be referenced.
    """

    key = param.String(default=None, constant=True, doc="""
        The name of the environment variable to observe.""")

    variable_type: ClassVar[str] = 'env'

    def __init__(self, **params):
        super().__init__(**params)
        key = self.key or self.name
        if key in os.environ:
            self.value = os.environ[key]


class Widget(Variable):
    """
    `Widget` variables dynamically reflect the current widget value.
    """

    kind = param.String(default=None, constant=True, doc="""
        The type of widget to instantiate.""")

    throttled = param.Boolean(default=True, constant=True, doc="""
        If the widget has a value_throttled parameter use that instead,
        ensuring that no intermediate events are generated, e.g. when
        dragging a slider.""")

    variable_type: ClassVar[str] = 'widget'

    _valid_keys: ClassVar[None] = None

    _required_keys = ["kind"]

    def __init__(self, **params):
        default = params.pop('default', None)
        refs = params.pop('refs', {})
        throttled = params.pop('throttled', True)
        label = params.pop('label', None)
        kind = params.pop('kind', None)
        widget = params.pop('widget', None)
        super().__init__(
            default=default, refs=refs, name=params.get('name'), label=label,
            throttled=throttled, kind=kind,
        )
        if '.' in self.kind:
            widget_type = resolve_module_reference(self.kind, _PnWidget)
        else:
            widget_type = getattr(pn.widgets, self.kind)
        if 'value' not in params and default is not None:
            params['value'] = default

        if self.label:
            params['name'] = self.label
        deserialized = {}
        for k, v in params.items():
            if k in widget_type.param:
                try:
                    v = widget_type.param[k].deserialize(v)
                except Exception:
                    pass
            deserialized[k] = v
        if widget:
            self._widget = widget
        else:
            self._widget = widget_type(**deserialized)
        self.value = self._widget.value
        if self.throttled and 'value_throttled' in self._widget.param:
            self._widget.link(self, value_throttled='value')
            self.param.watch(lambda e: self._widget.param.update({'value': e.new}), 'value')
        else:
            self._widget.link(self, value='value', bidirectional=True)

    def to_spec(self, context=None):
        """
        Exports the full specification to reconstruct this component.

        Returns
        -------
        Resolved and instantiated Component object
        """
        spec = super().to_spec(context=context)
        pvals = self._widget.param.values()
        for pname, pobj in self._widget.param.objects().items():
            pval = pvals[pname]
            if (
                pval is pobj.default or pname in spec or
                (pname == 'name' and not pval) or pname == 'value_throttled'
                or pname == 'design'  # Design is Panel 1.0 only
            ):
                continue
            try:
                # Gracefully handle failed equality checks
                if pval == pobj.default:
                    continue
            except Exception:
                pass
            spec[pname] = pvals[pname]
        return spec

    @property
    def panel(self):
        return self._widget


class Parameter(Variable):
    """
    `Parameter` variables reflect the current value of a parameter.
    """

    parameter = param.ClassSelector(class_=param.Parameter, constant=True,
        **disallow_refs, doc="""
        A parameter instance whose current value will be reflected
        on this variable.""")

    variable_type: ClassVar[str] = 'param'

    _allows_refs: ClassVar[bool] = False

    def __init__(self, **params):
        super().__init__(**params)
        self.parameter.owner.param.watch(
            lambda e: self.param.update(value=e.new),
            self.parameter.name
        )


class URLQuery(Variable):
    """
    `URLQuery` variables reflect the value of a URL query parameter.
    """

    key = param.String(default=None, constant=True, doc="""
        The URL query parameter to observe.""")

    variable_type: ClassVar[str] = 'url'

    def __init__(self, **params):
        super().__init__(**params)
        key = self.key or self.name
        if pn.state.location:
            pn.state.location.param.watch(self._update_value, 'search')
            self._update_value()
        if pn.state.session_args and key in pn.state.session_args:
            self.value = pn.state.session_args[key][0].decode('utf-8')

    def _update_value(self, *args):
        key = self.key or self.name
        self.value = pn.state.location.query_params.get(key, self.default)


class Cookie(Variable):
    """
    `Cookie` variables reflect the value of a cookie in the request.
    """

    key = param.String(default=None, constant=True, doc="""
        The cookie to observe.""")

    variable_type: ClassVar[str] = 'cookie'

    def __init__(self, **params):
        super().__init__(**params)
        if pn.state.cookies:
            key = self.key or self.name
            self.value = pn.state.cookies.get(key, self.default)


class UserInfo(Variable):
    """
    `UserInfo` variables reflect a value in the user info returned by an OAuth provider.
    """

    key = param.String(default=None, constant=True, doc="""
        The key in the OAuth pn.state.user_info dictionary to observe.""")

    variable_type: ClassVar[str] = 'user'

    def __init__(self, **params):
        super().__init__(**params)
        if pn.state.user_info:
            key = self.key or self.name
            self.value = pn.state.user_info.get(key, self.default)


class Header(Variable):
    """
    `Header` variables reflect the value of a request header.
    """

    key = param.String(default=None, constant=True, doc="""
        The request header to observe.""")

    variable_type: ClassVar[str] = 'header'

    def __init__(self, **params):
        super().__init__(**params)
        if pn.state.headers:
            key = self.key or self.name
            self.value = pn.state.headers.get(key, self.default)



from __future__ import annotations

import argparse
import ast
import os
import sys

import bokeh.command.util  # type: ignore

from bokeh.application.handlers.code import CodeHandler  # type: ignore
from bokeh.command.util import die  # type: ignore
from panel.command import Serve, transform_cmds
from panel.io.server import Application

from ..config import config

SOURCE_CODE = """
import pathlib

import lumen as lm
import lumen.ai as lmai
import panel as pn

from lumen.sources.duckdb import DuckDBSource

pn.extension("tabulator", "codeeditor", inline=False, template="fast")

llm = lmai.llm.OpenAI()

{table_initializer}

assistant = lmai.Assistant(
    llm=llm,
    agents=[
        lmai.agents.SourceAgent,
        lmai.agents.TableAgent,
        lmai.agents.TableListAgent,
        lmai.agents.SQLAgent,
        lmai.agents.PipelineAgent,
        lmai.agents.hvPlotAgent,
        lmai.agents.ChatAgent,
    ],
)
assistant.servable("Lumen.ai")
assistant.controls().servable(area="sidebar")
"""


class AIHandler(CodeHandler):
    ''' Modify Bokeh documents by using Lumen AI on a dataset.

    '''

    def __init__(self, *args, **kwargs):
        '''

        Keywords:
            filename (str) : a path to a dataset
            no_data (bool) : if True, do not load data

        '''
        table_initializer = ""
        if 'filename' in kwargs:
            table = os.path.abspath(kwargs['filename'])
            if table.endswith(".parq") or table.endswith(".parquet"):
                table = f"read_parquet('{table}')"
            elif table.endswith(".csv"):
                table = f"read_csv('{table}')"
            elif table.endswith(".json"):
                table = f"read_json_auto('{table}')"
            else:
                raise ValueError('Unsupported file format. Please provide a .parq, .parquet, .csv, or .json file.')

            table_initializer = f"""
lmai.memory["current_source"] = DuckDBSource(
    tables=["{table}"],
    uri=":memory:",
)
            """

        if 'no_data' in kwargs:
            kwargs.pop('no_data')
            kwargs['filename'] = 'no_data'

        kwargs['source'] = SOURCE_CODE.format(table_initializer=table_initializer)
        super().__init__(*args, **kwargs)


def build_single_handler_application(path: str | None, argv):

    if path is None or not os.path.isfile(path):
        handler = AIHandler(no_data=True)
    else:
        handler = AIHandler(filename=path)

    if handler.failed:
        raise RuntimeError("Error loading %s:\n\n%s\n%s " % (path, handler.error, handler.error_detail))

    application = Application(handler)
    return application

def build_single_handler_applications(paths: list[str], argvs: dict[str, list[str]] | None = None) -> dict[str, Application]:
    ''' Custom to allow for standalone `lumen-ai` command to launch without data'''
    applications: dict[str, Application] = {}
    argvs = argvs or {}

    if 'no_data' in sys.argv:
        application = build_single_handler_application(None, [])
        applications['/'] = application
    else:
        for path in paths:
            application = build_single_handler_application(path, argvs.get(path, []))

            route = application.handlers[0].url_path()

            if not route:
                if '/' in applications:
                    raise RuntimeError(f"Don't know the URL path to use for {path}")
                route = '/'
            applications[route] = application

    return applications

bokeh.command.util.build_single_handler_application = build_single_handler_application
bokeh.command.subcommands.serve.build_single_handler_applications = build_single_handler_applications

def main(args=None):
    start, template_vars = None, None
    for i, arg in enumerate(sys.argv):
        if '--template-vars' in arg:
            start = i
            if '=' in arg:
                end = i
                template_vars = arg.split('=')[1]
            else:
                end = i + 1
                template_vars = sys.argv[end]
            break

    if start is not None:
        sys.argv = sys.argv[:start] + sys.argv[end + 1:]
        config.template_vars = ast.literal_eval(template_vars)

    parser = argparse.ArgumentParser(
        prog="lumen-ai",
        description="""
        Lumen AI - Launch Lumen AI applications easily.\n\n To start the application without any
          data, simply run 'lumen-ai' with no additional arguments. You can upload data through
          the chat interface afterwards.
          """,
        epilog="See '<command> --help' to read about a specific subcommand."
    )

    parser.add_argument(
        '-v', '--version', action='version', version='Lumen AI 1.0.0'
    )

    subs = parser.add_subparsers(help="Sub-commands")

    serve_parser = subs.add_parser(Serve.name, help=
                                   """
                                   Run a bokeh server to serve the Lumen AI application.
                                   This command should be followed by dataset paths or directories
                                   to add to the chat memory, which can be a .parq, .parquet, .csv,
                                   or .json file. run `lumen-ai serve --help` for more options)
                                   """)
    serve_command = Serve(parser=serve_parser)
    serve_parser.set_defaults(invoke=serve_command.invoke)

    if len(sys.argv) > 1 and sys.argv[1] in ('--help', '-h'):
        args = parser.parse_args(sys.argv[1:])
        args.invoke(args)
        sys.exit()

    if len(sys.argv) == 1:
    # If no command is specified, start the server with an empty application
        sys.argv.extend(['serve', 'no_data'])

    sys.argv = transform_cmds(sys.argv)
    args = parser.parse_args(sys.argv[1:])

    try:
        ret = args.invoke(args)
    except Exception as e:
        die("ERROR: " + str(e))

    if ret is False:
        sys.exit(1)
    elif ret is not True and isinstance(ret, int) and ret != 0:
        sys.exit(ret)


if __name__ == "__main__":
    main()



import datetime as dt

from lumen.transforms.sql import (
    SQLColumns, SQLDistinct, SQLFilter, SQLGroupBy, SQLLimit, SQLMinMax,
)


def test_sql_group_by_single_column():
    assert (
        SQLGroupBy.apply_to('SELECT * FROM TABLE', by=['A'], aggregates={'AVG': 'B'}) ==
        """SELECT\n    A, AVG(B) AS B\nFROM ( SELECT * FROM TABLE )\nGROUP BY A"""
    )

def test_sql_group_by_multi_columns():
    assert (
        SQLGroupBy.apply_to('SELECT * FROM TABLE', by=['A'], aggregates={'AVG': ['B', 'C']}) ==
        """SELECT\n    A, AVG(B) AS B, AVG(C) AS C\nFROM ( SELECT * FROM TABLE )\nGROUP BY A"""
    )

def test_sql_limit():
    assert (
        SQLLimit.apply_to('SELECT * FROM TABLE', limit=10) ==
        """SELECT\n    *\nFROM ( SELECT * FROM TABLE )\nLIMIT 10"""
    )

def test_sql_columns():
    assert (
        SQLColumns.apply_to('SELECT * FROM TABLE', columns=['A', 'B']) ==
        """SELECT\n    "A", "B"\nFROM ( SELECT * FROM TABLE )"""
    )

def test_sql_distinct():
    assert (
        SQLDistinct.apply_to('SELECT * FROM TABLE', columns=['A', 'B']) ==
        """SELECT DISTINCT\n    "A", "B"\nFROM ( SELECT * FROM TABLE )"""
    )

def test_sql_min_max():
    assert (
        SQLMinMax.apply_to('SELECT * FROM TABLE', columns=['A', 'B']) ==
        """SELECT\n    MIN("A") as "A_min", MAX("A") as "A_max", MIN("B") as "B_min", MAX("B") as "B_max"\nFROM ( SELECT * FROM TABLE )"""
    )

def test_sql_filter_none():
    assert (
        SQLFilter.apply_to('SELECT * FROM TABLE', conditions=[('A', None)]) ==
        """SELECT\n    *\nFROM ( SELECT * FROM TABLE )\nWHERE ( A IS NULL )"""
    )

def test_sql_filter_scalar():
    assert (
        SQLFilter.apply_to('SELECT * FROM TABLE', conditions=[('A', 1)]) ==
        """SELECT\n    *\nFROM ( SELECT * FROM TABLE )\nWHERE ( A = 1 )"""
    )


def test_sql_filter_isin():
    assert (
        SQLFilter.apply_to('SELECT * FROM TABLE', conditions=[('A', ['A', 'B', 'C'])]) ==
        """SELECT\n    *\nFROM ( SELECT * FROM TABLE )\nWHERE ( A IN ('A', 'B', 'C') )"""
    )

def test_sql_filter_datetime():
    assert (
        SQLFilter.apply_to('SELECT * FROM TABLE', conditions=[('A', dt.datetime(2017, 4, 14))]) ==
        """SELECT\n    *\nFROM ( SELECT * FROM TABLE )\nWHERE ( A = '2017-04-14 00:00:00' )"""
    )

def test_sql_filter_date():
    assert (
        SQLFilter.apply_to('SELECT * FROM TABLE', conditions=[('A', dt.date(2017, 4, 14))]) ==
        """SELECT\n    *\nFROM ( SELECT * FROM TABLE )\nWHERE ( A BETWEEN '2017-04-14 00:00:00' AND '2017-04-14 23:59:59' )"""
    )

def test_sql_filter_date_range():
    assert (
        SQLFilter.apply_to('SELECT * FROM TABLE', conditions=[('A', (dt.date(2017, 2, 22), dt.date(2017, 4, 14)))]) ==
        """SELECT\n    *\nFROM ( SELECT * FROM TABLE )\nWHERE ( A BETWEEN '2017-02-22 00:00:00' AND '2017-04-14 23:59:59' )"""
    )

def test_sql_filter_datetime_range():
    assert (
        SQLFilter.apply_to('SELECT * FROM TABLE', conditions=[('A', (dt.datetime(2017, 2, 22), dt.datetime(2017, 4, 14)))]) ==
        """SELECT\n    *\nFROM ( SELECT * FROM TABLE )\nWHERE ( A BETWEEN '2017-02-22 00:00:00' AND '2017-04-14 00:00:00' )"""
    )
